
/////////////////////////////////////////////////////////////////////////////////////
// GameManager.cpp
/////////////////////////////////////////////////////////////////////////////////////

#include "nds.h"
#include <stdio.h>

#include "Viewport.h"
#include "SpriteManager.h"
#include "Car.h"
#include "Road.h"
#include "HUD.h"
#include "GameManager.h"
#include "inflate.h"

// the graphics
#include "titleScreen_bin.h"
#include "titleScreen_pal.h"
#include "sprites_bin.h"
#include "sprites_pal.h"

// sin table
#include "sin_bin.h"

/////////////////////////////////////////////////////////////////////////////////////
// variables
/////////////////////////////////////////////////////////////////////////////////////

// the background horizontal scroll value (24.8 FP)
static int backgroundFPX;

/////////////////////////////////////////////////////////////////////////////////////
// GameManager()
/////////////////////////////////////////////////////////////////////////////////////

GameManager::GameManager(void) {

	// set the default values
	viewport = new Viewport();
	viewport->reshape(0, 0, 256, 192);

	// create the road
	road = new Road();

	// create the spriteManager
	spriteManager = new SpriteManager();

	// create the HUD
	hUD = new HUD();

	// create the player car
	carPlayer = new Car(CAR_TYPE_PLAYER);
}

/////////////////////////////////////////////////////////////////////////////////////
// ~GameManager()
/////////////////////////////////////////////////////////////////////////////////////

GameManager::~GameManager(void) {

	if (road != NULL)
		delete road;
	if (spriteManager != NULL)
		delete spriteManager;
}

/////////////////////////////////////////////////////////////////////////////////////
// init() - init the hardware to run the game
/////////////////////////////////////////////////////////////////////////////////////

void GameManager::init(void) {

	// TODO: by adding abstraction here we could make this game platform independent,
	// but that's too much work for a showcase assignment

	// setup GFX-A (BG2 ~ affine extended BG)
	videoSetMode(MODE_5_2D | DISPLAY_BG2_ACTIVE);
	BG2_CR = BG_BMP8_256x256;

	// give it BANK A (128KB) VRAM
	vramSetBankA(VRAM_A_MAIN_BG);

	// reset affine transformations
	BG2_XDX = 1 << 8;
	BG2_XDY = 0;
	BG2_YDX = 0;
	BG2_YDY = 1 << 8;
	BG2_CX = 0;
	BG2_CY = 0 << 8;

	// setup GFX-B (BG2 ~ affine extended BG)
	videoSetModeSub(MODE_5_2D | DISPLAY_BG2_ACTIVE);
	SUB_BG2_CR = BG_BMP8_256x256;

	// give it BANK C (128KB) VRAM
	vramSetBankC(VRAM_C_SUB_BG);

	// reset affine transformations
	SUB_BG2_XDX = 1 << 8;
	SUB_BG2_XDY = 0;
	SUB_BG2_YDX = 0;
	SUB_BG2_YDY = 1 << 8;
	SUB_BG2_CX = 0;
	SUB_BG2_CY = 0 << 8;

	// copy the images
	unsigned short *out = (unsigned short *)0x6000000;
	unsigned short *in = (unsigned short *)titleScreen_bin;
	inflate((u8 *)in, out);

	// load sprites
	spriteManager->spriteSheetLoad((unsigned char *)sprites_bin, (unsigned short *)sprites_pal);

	// copy the palettes
	// TODO: implement fades?
	in = (unsigned short *)titleScreen_pal;
	for (int i = 0; i < 256; i++)
		BG_PALETTE[i] = in[i];

	// flush the CPU caches, to make that data really go into vram
	DC_FlushAll();
}

/////////////////////////////////////////////////////////////////////////////////////
// reset() - reset the game
/////////////////////////////////////////////////////////////////////////////////////

void GameManager::reset(void) {

	carPlayer->reset();
	road->reset();

	// reset the lap statistics
	lapTimeCurrent = 0;
	lapTimeBest = 0;
	lap = 1;

	// reset the background
	backgroundFPX = 0;
}

/////////////////////////////////////////////////////////////////////////////////////
// precalculateScreen() - set the view size, and do some precalculations
/////////////////////////////////////////////////////////////////////////////////////

void GameManager::precalculateScreen(int viewDX, int viewDY) {

	// center the view on DS's 256x192 screen
	viewport->reshape((256 - viewDX) >> 1,
										(192 - viewDY) >> 1,
										viewDX,
										viewDY);

	// reconstruct the road scanlines
	road->precalculateRoad(viewDX, viewDY);
}

/////////////////////////////////////////////////////////////////////////////////////
// iteration() - the main loop iteration of the game
/////////////////////////////////////////////////////////////////////////////////////

void GameManager::iteration(int deltaTime, int keysCurrent, int keysOld) {

	// NOTE: on some systems the deltaTime here wouldn't always be 1000/60,
	// and because of this we use a variable here instead of the hard value

	// handle acceleration and deacceleration
	int maxSpeed = CAR_MAX_SPEED;
	if (keysCurrent & KEY_UP)
		carPlayer->speed += deltaTime / 6;
	else if (keysCurrent & KEY_DOWN)
		carPlayer->speed -= deltaTime / 5;

	if (carPlayer->speed > maxSpeed)
		carPlayer->speed = maxSpeed;

	// driving on grass?
	bool onGrass = false;
	int carX = (viewport->startX << 1) + viewport->dX;
	if (carX < road->roadStartX[0] || carX > road->roadEndX[0]) {
		maxSpeed = maxSpeed*50/100;
		onGrass = true;
	}

#ifdef DEBUG
	// change the sky color to indicate the we are driving off road
	if (onGrass)
		BG_PALETTE_SUB[SPRITE_COLOR_SKY] = RGB15(147 >> 3, 40 >> 3, 40 >> 3);
	else
		BG_PALETTE_SUB[SPRITE_COLOR_SKY] = RGB15(98 >> 3, 147 >> 3, 231 >> 3);
#endif

	if (carPlayer->speed > maxSpeed)
		carPlayer->speed -= deltaTime / 5;
	if (carPlayer->speed < 0)
		carPlayer->speed = 0;

	// handle turning (the faster the car moves the faster it turns)
	// depends also on the width of the screen
	int turnSpeed = deltaTime * carPlayer->speed * 60 / 1000 * viewport->dX;
	int moveFPX = 0;
	turnSpeed >>= 7;
	if (keysCurrent & KEY_LEFT)
		moveFPX -= turnSpeed;
	else if (keysCurrent & KEY_RIGHT)
		moveFPX += turnSpeed;

	// add sliding (depends on the road's turning angle, the player car's speed, and on the width of the screen)
	int angle = road->getAngle(carPlayer);
	turnSpeed = angle * deltaTime * carPlayer->speed;
	turnSpeed >>= 14;
	turnSpeed = (turnSpeed * viewport->dX) >> 7;
	moveFPX -= turnSpeed;

	// don't let the player drive too far away from the road
	if (carX < road->roadStartX[0] - 30 && moveFPX < 0)
		moveFPX = 0;
	else if (carX > road->roadEndX[0] + 30 && moveFPX > 0)
		moveFPX = 0;

	carPlayer->fPX += moveFPX;

	// move the background
	int spriteDX = spriteManager->getSpriteWidth(SPRITE_ID_BACKGROUND_CITY) << 8;
	backgroundFPX -= (angle * deltaTime * carPlayer->speed) >> 12;
	if (backgroundFPX <= -spriteDX >> 1)
		backgroundFPX += spriteDX;
	else if (backgroundFPX >= spriteDX >> 1)
		backgroundFPX -= spriteDX;

	// move the car forward
	carPlayer->fPY += carPlayer->speed * deltaTime * 60 / 1000;

	// animate the car (the faster it moves the faster is the animation)
	carPlayer->frameTimer += deltaTime * carPlayer->speed;

	// run the timers
	lapTimeCurrent += deltaTime;

	if (lap == 1)
		lapTimeBest = lapTimeCurrent;

	// manage the road as the player car drives on
	int newLap = road->update(carPlayer);

	// drove to a new lap?
	if (lap != newLap) {
		lap = newLap;
		if (lapTimeBest > lapTimeCurrent)
			lapTimeBest = lapTimeCurrent;
		lapTimeCurrent = 0;
	}
}

/////////////////////////////////////////////////////////////////////////////////////
// draw() - draw the screen
/////////////////////////////////////////////////////////////////////////////////////

void GameManager::draw(void) {

	volatile unsigned short *vram = (volatile unsigned short *)0x6200000;
	volatile unsigned int *ptr32;
	int i, j;

	// TODO: doing the drawing this way is suboptimal on DS, but it's quick to code.
	// if done properly this code would use the 2D and 3D hardware acceleration.
	// anyway, the purpose of this project is to analyze my programming skills, I guess, not to
	// create resellable code.

	// clear the background (256x192)

	// sky first
	ptr32 = (volatile unsigned int *)vram;
	j = (SPRITE_COLOR_SKY << 24) | (SPRITE_COLOR_SKY << 16) | (SPRITE_COLOR_SKY << 8) | SPRITE_COLOR_SKY;
	for (i = 0; i < 256*192/2/4; i += 8) {
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
	}

	// draw the HUD
	// we can do this here as nothing is going to be drawn on top of the HUD
	// this way we can skip implementing double buffered drawing :D
	hUD->draw(vram, spriteManager, viewport, lap, lapTimeCurrent, lapTimeBest, (carPlayer->speed << 8) / CAR_MAX_SPEED);

	// next the background city in the horizon
	// the background is drawn from three identical pieces that loop
	int spriteDX = spriteManager->getSpriteWidth(SPRITE_ID_BACKGROUND_CITY) >> 1;
	int centerX = viewport->startX + (viewport->dX >> 1) + (backgroundFPX >> 8) - spriteDX;
	if (backgroundFPX > 0)
		centerX -= spriteDX;
	for (j = 0; j < 3; j++)
		spriteManager->drawSprite(vram, SPRITE_ID_BACKGROUND_CITY, centerX + (j * spriteDX) , viewport->startY + (viewport->dY >> 1), SPRITE_PIVOT_BOTTOM, viewport);

	// then the ground
	j = (SPRITE_COLOR_GROUND << 24) | (SPRITE_COLOR_GROUND << 16) | (SPRITE_COLOR_GROUND << 8) | SPRITE_COLOR_GROUND;
	for (; i < 256*192/4; i += 8) {
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
		*(ptr32++) = j;
	}

	// draw the road
	road->draw(vram, carPlayer, spriteManager, viewport);

	// draw the player car
	int frameID = SPRITE_ID_PLAYER_CAR_FRAME_1;
	i = carPlayer->frameTimer / 2000;
	if (i & 1)
		frameID = SPRITE_ID_PLAYER_CAR_FRAME_2;

	spriteManager->drawSprite(vram, frameID, viewport->startX + (viewport->dX >> 1), viewport->startY + viewport->dY, SPRITE_CENTERED_X | SPRITE_PIVOT_BOTTOM, viewport);

	// flush the data cache so that all the changes will actually become visible
	DC_FlushAll();
}
