
/////////////////////////////////////////////////////////////////////////////////////
// Road.cpp
/////////////////////////////////////////////////////////////////////////////////////

#include "nds.h"
#include <stdio.h>

#include "Viewport.h"
#include "SpriteManager.h"
#include "Car.h"
#include "Road.h"

// sin table
#include "sin_bin.h"

/////////////////////////////////////////////////////////////////////////////////////
// internal variables
/////////////////////////////////////////////////////////////////////////////////////

// the race track
// NOTE: if this game had more than one level, this only level wouldn't be
// here hardcoded as it would come from an editor...

// the number of segments the race track has
static const int levelSegments = 22;

// the turn info
static const int levelDataTurn[] = {
	0, 0, 1, 0, 0, 0, 1, 0, 0, -1, 0, 1, 0, 1, 0, 0, 1, 0, -1, 0, 1, 0
};

// the segment length
// TODO: make this an array
static const int levelSegmentLength = 64;

// the last segment we unpacked
static int levelSegmentLastUnpacked;

// the roadside ads
#define SPRITES_ADS 6

static const int spritesAds[] = {
	SPRITE_ID_ROADSIDE_AD_1,
	SPRITE_ID_ROADSIDE_AD_2,
	SPRITE_ID_ROADSIDE_AD_3,
	SPRITE_ID_ROADSIDE_AD_4,
	SPRITE_ID_ROADSIDE_AD_5,
	SPRITE_ID_ROADSIDE_AD_6,
};

/////////////////////////////////////////////////////////////////////////////////////
// Road()
/////////////////////////////////////////////////////////////////////////////////////

Road::Road(void) {

	// init
	roadDX = NULL;
	roadStartX = NULL;
	roadEndX = NULL;
	roadPointers = NULL;
	roadTurn = new int[ROAD_SLICES];
	roadBits = new unsigned char[ROAD_SLICES];
}

/////////////////////////////////////////////////////////////////////////////////////
// ~Road()
/////////////////////////////////////////////////////////////////////////////////////

Road::~Road(void) {

	if (roadDX != NULL)
		delete[] roadDX;
	if (roadStartX != NULL)
		delete[] roadStartX;
	if (roadEndX != NULL)
		delete[] roadEndX;
	if (roadPointers != NULL)
		delete[] roadPointers;
	if (roadTurn != NULL)
		delete[] roadTurn;
	if (roadBits != NULL)
		delete[] roadBits;
}

/////////////////////////////////////////////////////////////////////////////////////
// precalculateRoad() - precalculate the road
/////////////////////////////////////////////////////////////////////////////////////

void Road::precalculateRoad(int viewDX, int viewDY) {

	// bottom road scanline is 140% of the view width, top is 5%
	int bottomDX = (viewDX << 8)*140/100;
	int topDX = (viewDX << 8)*5/100;

	// make sure the topmost road scanline is always wider than three pixels
	if (topDX < (4 << 8))
		topDX = 4 << 8;

	// the road height is always 50% of the viewport height
	viewDY >>= 1;

	int dX = bottomDX - topDX;
	int addDX = -dX / (viewDY - 1);

	if (roadDX != NULL)
		delete[] roadDX;
	if (roadStartX != NULL)
		delete[] roadStartX;
	if (roadEndX != NULL)
		delete[] roadEndX;
	if (roadPointers != NULL)
		delete[] roadPointers;

	roadDX = new unsigned int[viewDY];
	roadStartX = new int[viewDY];
	roadEndX = new int[viewDY];
	roadPointers = new int[viewDY];

	// precalculate the road scanline widths
	// TODO: move more stuff here from Road::draw(), if possible
	dX = bottomDX;
	for (int i = 0; i < viewDY; i++) {
		roadDX[i] = dX;
		dX += addDX;
	}
}

/////////////////////////////////////////////////////////////////////////////////////
// getAngle() - get the angle of the road under the specified car
/////////////////////////////////////////////////////////////////////////////////////

int Road::getAngle(Car *car) {

	short *sinTable = (short *)sin_bin;
	int roadPointer = (car->fPY >> 8) & (ROAD_SLICES - 1);
	return sinTable[roadTurn[roadPointer] & 511];
}

/////////////////////////////////////////////////////////////////////////////////////
// draw() - draw the road
/////////////////////////////////////////////////////////////////////////////////////

void Road::draw(volatile unsigned short *vram, Car *carPlayer, SpriteManager *spriteManager, Viewport *viewport) {

	short *sinTable = (short *)sin_bin;
	int j = 2 << 8;
	int addJ = 2 << 8;
	int addAddJ = 128-16;
	int screenY = viewport->startY + viewport->dY - 1;
	int angle = 0;
	int angleAccum = 0;
	volatile unsigned short *ptr16;

	// draw the road one scanline at a time, from bottom to top
	for (int i = 0; i < viewport->dY >> 1; i++) {
		int roadPointer = ((carPlayer->fPY + (j >> 8)) >> 8) & (ROAD_SLICES - 1);
		roadPointers[i] = roadPointer;

		// determine road color
		int color;
		if (roadPointer & 1)
			color = SPRITE_COLOR_ROAD_1;
		else
			color = SPRITE_COLOR_ROAD_2;

		// finish line?
		if (roadBits[roadPointer] & ROAD_BIT_FINISH_LINE)
			color = SPRITE_COLOR_FINISH_LINE;

		color = (color << 8) | color;

		// calculate the center of the road
		int centerX = (viewport->startX << 1) + viewport->dX - (carPlayer->fPX >> 8);
		angle += sinTable[roadTurn[roadPointer] & 511];
		angleAccum += angle;
		centerX += angleAccum >> 14;

		// add perspective when moving left and right
		int dX = roadDX[i] >> 8;
		int tilt = (dX * carPlayer->fPX) >> 13;
		centerX -= tilt;

		int startX = centerX - (dX >> 1);
		int endX = centerX + (dX >> 1);

		roadStartX[i] = startX;
		roadEndX[i] = endX;

		// X clip
		if (startX < viewport->startX << 1)
			startX = viewport->startX << 1;
		if (endX > (viewport->startX << 1) + (viewport->dX << 1))
			endX = (viewport->startX << 1) + (viewport->dX << 1);

		// fill the scanline
		// TODO: make this pixel perfect. currently our horizontal resolution is calculated in pixel pairs...
		// but this works for now
		// TODO: make this behind something like Graphics.fillHorizontalLine() etc...
		ptr16 = &vram[(screenY << 7) + (startX >> 1)];
		for (int k = startX >> 1; k < endX >> 1; k++)
			*(ptr16++) = color;

		j += addJ;
		addJ += addAddJ;
		addAddJ += 8;
		screenY--;
	}

	// next draw all objects, top to bottom as the objects are filled from bottom to top
	screenY = viewport->startY + (viewport->dY >> 1);
	int previousBits = -1;
	int reverseIndex = (viewport->dY >> 1) - 1;
	for (int i = 0; i < viewport->dY >> 1; i++) {
		int roadPointer = roadPointers[reverseIndex];
		int bits = roadBits[roadPointer];
		int startX = roadStartX[reverseIndex];
		int endX = roadEndX[reverseIndex];

		// draw a roadside ad?
		if ((bits & ROAD_BIT_ROADSIDE_AD) && bits != previousBits) {
			// yes!
			int spriteID = i / ((viewport->dY >> 1) / 5);

			int posX = endX + (roadDX[reverseIndex] >> 11);
			spriteManager->drawSprite(vram, spritesAds[spriteID], posX >> 1, screenY, SPRITE_CENTERED_X | SPRITE_PIVOT_BOTTOM, viewport);
			posX = startX - (roadDX[reverseIndex] >> 11);
			spriteManager->drawSprite(vram, spritesAds[spriteID], posX >> 1, screenY, SPRITE_CENTERED_X | SPRITE_PIVOT_BOTTOM, viewport);
		}

		// store the bits for the next iteration
		previousBits = bits;

		screenY++;
		reverseIndex--;
	}
}

/////////////////////////////////////////////////////////////////////////////////////
// reset() - reset the road / level
/////////////////////////////////////////////////////////////////////////////////////

void Road::reset(void) {

	int i;

	// reset the road
	for (i = 0; i < ROAD_SLICES; i++) {
		roadTurn[i] = 0;
		roadBits[i] = 0;
	}

	levelSegmentLastUnpacked = -1;

	// unpack a couple of road segments
	unpack(0);
	unpack(1);
}

/////////////////////////////////////////////////////////////////////////////////////
// update() - update the road ringbuffers
/////////////////////////////////////////////////////////////////////////////////////

int Road::update(Car *carPlayer) {

	// get the current segment
	int current = (carPlayer->fPY >> 8) / levelSegmentLength;

	// unpack a segment two segments away from the current position
	int target = current + 2;

	if (levelSegmentLastUnpacked != target)
		unpack(target);

	return (current / levelSegments) + 1;
}

/////////////////////////////////////////////////////////////////////////////////////
// unpack() - unpack the specified road segment into the road ringbuffers
/////////////////////////////////////////////////////////////////////////////////////

void Road::unpack(int segment) {

	levelSegmentLastUnpacked = segment;

	// loop the track
	segment %= levelSegments;

	int out = levelSegmentLastUnpacked * levelSegmentLength;
	int in = levelDataTurn[segment];

	for (int i = 0; i < levelSegmentLength; i++) {
		// unpack the piece
		out &= (ROAD_SLICES - 1);
		roadTurn[out] = in;

		// at the center of every segment we have a roadside ad
		if (i == levelSegmentLength >> 1)
			roadBits[out] = ROAD_BIT_ROADSIDE_AD;
		else
			roadBits[out] = 0;

		out++;
	}

	// add finish line?
	if (segment == 0) {
		// yes, make it thick
		out = levelSegmentLastUnpacked * levelSegmentLength;
		out &= (ROAD_SLICES - 1);
		roadBits[out++] = ROAD_BIT_FINISH_LINE;
		out &= (ROAD_SLICES - 1);
		roadBits[out++] = ROAD_BIT_FINISH_LINE;
		out &= (ROAD_SLICES - 1);
		roadBits[out++] = ROAD_BIT_FINISH_LINE;
	}
}
